# coding=utf-8
"""Tree synchronization

Parsing of syncing options

"""

import os
import shutil
import threading
import time

from subprocess import Popen, PIPE

from soundforest.defaults import SOUNDFOREST_USER_DIR
from soundforest.cli import ScriptThread, ScriptThreadManager
from soundforest.config import ConfigDB
from soundforest.log import SoundforestLogger
from soundforest.tree import Tree, Track, TreeError

RSYNC_DELETE_FLAGS = (
    '--del',
    '--delete',
    '--delete-before',
    '--delete-during',
    '--delete-after',
    '--delete-delay',
    '--delete-excluded'
)
DEFAULT_DELETE_FLAG = '--delete-before'


class SyncError(Exception):
    pass


def ntfs_rename(path):
    REPLACE_MAP = {
        '|': '-',
        '>': '-',
        '<': '-',
        '"': '',
        ':': ' - ',
        '?': '',
        '!': '',
        '*': '',
    }
    for c, r in REPLACE_MAP.items():
        path = path.replace(c, r)

    # Silly system does not allow components ending with .
    path = os.sep.join(x.rstrip('. ') for x in path.split(os.sep))
    return path

RENAME_CALLBACKS = {
    'ntfs': ntfs_rename,
}


class SyncThread(ScriptThread):
    def __init__(self, manager, index, src, dst, delete=False):
        super(SyncThread, self).__init__('sync')
        self.manager = manager
        self.index = index
        self.delete = delete

        if isinstance(src, Tree):
            self.src_tree = src
            self.src = src.path

        elif isinstance(src, basestring):
            self.src_tree = Tree(src)
            self.src = os.path.expandvars(src).rstrip(os.sep)

        else:
            raise SyncError('Src is not string or Tree object: {0}'.format(src))

        if isinstance(dst, Tree):
            self.dst_tree = dst
            self.dst = dst.path

        elif isinstance(dst, basestring):
            self.dst_tree = Tree(dst)
            self.dst = os.path.expandvars(dst).rstrip(os.sep)

        else:
            raise SyncError('Dst is not string or Tree object: {0}'.format(dst))

    def run(self):
        raise NotImplementedError('Must be implemented in inheriting class')


class FilesystemSyncThread(SyncThread):
    def __init__(self, manager, index, src, dst, delete=False, rename=None):
        super(FilesystemSyncThread, self).__init__(manager, index, src, dst, delete)

        if rename is not None:
            try:
                rename = RENAME_CALLBACKS[rename]
            except KeyError:
                raise SyncError('Unknown rename callback: {0}'.format(rename))

        self.rename = rename

    def copy_track(self, src, dst):
        try:
            shutil.copyfile(src, dst)

        except IOError, (ecode, emsg):
            raise SyncError('Error writing to {0}: {1}'.format(dst, emsg))

        except OSError, (ecode, emsg):
            raise SyncError('Error writing to {0}: {1}'.format(dst, emsg))

    def run(self):
        if not os.path.isdir(self.src_tree.path):
            raise SyncError('Source not available while syncing: {0}'.format(self.src_tree.path))

        if not os.path.isdir(self.dst_tree.path):
            raise SyncError('Destination not available while syncing: {0}'.format(self.dst_tree.path))

        src = self.src_tree
        dst = self.dst_tree
        i=0

        for album in src.as_albums():
            dst_album_path = os.path.join(dst.path, src.relative_path(album.path))
            if self.rename is not None:
                dst_album_path = self.rename(dst_album_path)

            if not os.path.isdir(dst_album_path):
                try:
                    self.log.debug('Create directory: {0}'.format(dst_album_path))
                    os.makedirs(dst_album_path)

                except OSError, (ecode, emsg):
                    self.log.info('Error creating directory {0}: {1}'.format(dst_album_path, emsg))
                    continue

            for track in album:
                i+=1
                dst_track_path = os.path.join(dst.path, track.relative_path())

                if self.rename:
                    dst_track_path = self.rename(dst_track_path)
                dst_track = Track(os.path.join(dst_track_path))

                modified = False
                if not os.path.isfile(dst_track.path):
                    self.log.info('{0:6d} new: {1}'.format(i, dst_track.path))
                    modified = True

                elif track.size != dst_track.size:
                    self.log.info('{0:6d} modified: {1}'.format(i, dst_track.path))
                    modified = True

                if modified:
                    try:
                        self.copy_track(track.path, dst_track.path)

                    except SyncError, emsg:
                        print emsg
                        continue


class RsyncThread(SyncThread):
    def __init__(self, manager, index, src, dst, flags, delete=False):
        super(RsyncThread, self).__init__(manager, index, src, dst, delete)
        if isinstance(flags, basestring):
            flags = flags.split()

        if delete and not RSYNC_DELETE_FLAGS.intersection(set(flags)):
            flags.insert(0, DEFAULT_DELETE_FLAG)

        self.flags = flags

    def run(self):
        command = ['rsync', '-av'] + self.flags + ['{0}/'.format(self.src), '{0}/'.format(self.dst)]
        print 'running ', ' '.join(command)

        try:
            self.log.info('Running: {0}'.format(' '.join(command)))

            p = Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            rval = None
            while rval is None:
                while True:
                    l = p.stdout.readline()
                    if l == '': break
                    self.log.info(l.rstrip())
                time.sleep(0.2)
                rval = p.poll()

            if rval != 0:
                self.log.info('Error running command {0}: {1}'.format(self, p.stderr.read()))

        except KeyboardInterrupt:
            self.log.debug('Rsync interrupted')
            raise KeyboardInterrupt

        self.log.info('Finished: {0}'.format(' '.join(command)))


class SyncManager(ScriptThreadManager):
    def __init__(self, threads=None, delete=False, debug=False):
        super(SyncManager, self).__init__('sync', threads)
        self.delete = delete
        self.debug = debug

        if not debug:
            self.log = SoundforestLogger('sync').register_file_handler('sync', SOUNDFOREST_USER_DIR)
            SoundforestLogger('sync').set_level('INFO')

        else:
            self.log = SoundforestLogger().default_stream

    def parse_target(self, name):
        try:
            target = self.db.sync[name]

        except KeyError:
            return None

        if not 'src' in target:
            raise SyncError('Target missing source')

        if not 'dst' in target:
            raise SyncError('Target missing destination')

        return target

    @property
    def rename_callbacks(self):
        return RENAME_CALLBACKS

    def get_entry_handler(self, index, config):
        sync_type = config.pop('type', None)
        if sync_type=='rsync':
            return RsyncThread(manager=self, index=index, **config)

        elif sync_type=='directory':
            if 'flags' in config:
                del config['flags']
            return FilesystemSyncThread(manager=self, index=index, **config)

        else:
            raise SyncError('BUG: invalid sync type in thread config')

    def enqueue(self, config):
        if not isinstance(config, dict):
            raise SyncError('Enqueue requires a dictionary')

        sync_type = config.get('type', None)
        if sync_type not in ['rsync', 'directory']:
            raise SyncError('Unknown sync type in config: {0}'.format(sync_type))

        if 'delete' not in config:
            config['delete'] = self.delete

        for k in ('id', 'name', 'defaults'):
            if k in config:
                config.pop(k)

        self.append(config)

    def run(self):
        if len(self)==0:
            return

        total = len(self)
        while len(self)>0:
            active = threading.active_count()
            if active > self.threads:
                time.sleep(0.5)
                continue

            index = '{0:d}/{1:d}'.format(total-len(self)+1, total)
            t = self.get_entry_handler(index, self.pop(0))
            t.start()

        active = threading.active_count()
        while active > 1:
            time.sleep(0.5)
            active = threading.active_count()



# coding=utf-8
"""Logging functions

Soundforest debug and file logging facilities

"""
import os
import sys
import fnmatch
import re
import urllib
import bz2
import gzip
import logging
import logging.handlers

from datetime import datetime, timedelta

from soundforest import SoundforestError

DEFAULT_LOGFORMAT = '%(module)s %(levelname)s %(message)s'
DEFAULT_TIME_FORMAT = "%Y-%m-%d %H:%M:%S"
DEFAULT_LOGFILEFORMAT = '%(asctime)s %(module)s.%(funcName)s %(message)s'
DEFAULT_LOGSIZE_LIMIT = 2**20
DEFAULT_LOG_BACKUPS = 10

DEFAULT_SYSLOG_FORMAT = '%(message)s'
DEFAULT_SYSLOG_LEVEL =  logging.handlers.SysLogHandler.LOG_WARNING
DEFAULT_SYSLOG_FACILITY = logging.handlers.SysLogHandler.LOG_USER

# Mapping to set syslog handler levels via same classes as normal handlers
LOGGING_LEVEL_NAMES = ( 'DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL' )
SYSLOG_LEVEL_MAP = {
    logging.handlers.SysLogHandler.LOG_DEBUG:   logging.DEBUG,
    logging.handlers.SysLogHandler.LOG_NOTICE:  logging.INFO,
    logging.handlers.SysLogHandler.LOG_INFO:    logging.INFO,
    logging.handlers.SysLogHandler.LOG_WARNING: logging.WARN,
    logging.handlers.SysLogHandler.LOG_ERR:     logging.ERROR,
    logging.handlers.SysLogHandler.LOG_CRIT:    logging.CRITICAL,
}

# Local syslog device varies by platform
if sys.platform == 'linux2' or fnmatch.fnmatch(sys.platform, '*bsd*'):
    DEFAULT_SYSLOG_ADDRESS = '/dev/log'
elif sys.platform == 'darwin':
    DEFAULT_SYSLOG_ADDRESS = '/var/run/syslog'
else:
    DEFAULT_SYSLOG_ADDRESS = ( 'localhost', 514 )

class SoundforestLogger(object):
    """SoundforestLogger

    Class for common script logging tasks. Implemented as singleton to prevent
    errors in duplicate handler initialization.

    """
    __instances = {}
    def __init__(self, name=None):
        name = name is not None and name or self.__class__.__name__

        if not SoundforestLogger.__instances.has_key(name):
            SoundforestLogger.__instances[name] = SoundforestLogger.LoggerInstance(name)

        self.__dict__['_SoundforestLogger__instances'] = SoundforestLogger.__instances
        self.__dict__['name'] = name

    class LoggerInstance(dict):
        """LoggerInstance

        Singleton implementation of logging configuration for one program

        """
        def __init__(self, name):
            self.name = name
            self.loglevel = logging.Logger.root.level
            self.register_stream_handler('default_stream')

        def __getattr__(self, attr):
            if attr in self.keys():
                return self[attr]
            raise AttributeError('No such LoggerInstance attribute: {0}'.format(attr))

        def __get_or_create_logger__(self, name):
            if name not in self.keys():
                for logging_manager in logging.Logger.manager.loggerDict.values():
                    if hasattr(logging_manager, 'name') and logging_manager.name==name:
                        self[name] = logging.getLogger(name)
                        break

            if name not in self.keys():
                self[name] = logging.getLogger(name)

            return self[name]

        def __match_handlers__(self, handler_list, handler):
            def match_handler(a, b):
                if type(a) != type(b):
                    return False

                if isinstance(a, logging.StreamHandler):
                    for k in ('stream', 'name',):
                        if getattr(a, k) != getattr(b, k):
                            return False
                    return True

                if isinstance(a, logging.handlers.SysLogHandler):
                    for k in ( 'address', 'facility', ):
                        if getattr(a, k) != getattr(b, k):
                            return False
                    return True

                if isinstance(a, logging.handlers.HTTPHandler):
                    for k in ( 'host', 'url', 'method', ):
                        if getattr(a, k) != getattr(b, k):
                            return False
                    return True

                return True

            if not isinstance(handler, logging.Handler):
                raise LoggerError('Not an instance of logging.Handler: {0}'.format(handler))

            if not isinstance(handler_list, list):
                raise LoggerError('BUG handler_list must be a list instance')

            for match in handler_list:
                if match_handler(match, handler):
                    return True

            return False

        def register_stream_handler(self, name, logformat=None, timeformat=None):
            """
            Register a common log stream handler
            """

            if logformat is None:
                logformat = DEFAULT_LOGFORMAT
            if timeformat is None:
                timeformat = DEFAULT_TIME_FORMAT

            logger = self.__get_or_create_logger__(name)
            handler = logging.StreamHandler()
            if self.__match_handlers__(self.default_stream.handlers, handler):
                return

            if not self.__match_handlers__(logger.handlers, handler):
                handler.setFormatter(logging.Formatter(logformat, timeformat))
                logger.addHandler(handler)

        def register_syslog_handler(self, name,
                address=DEFAULT_SYSLOG_ADDRESS,
                facility=DEFAULT_SYSLOG_FACILITY,
                default_level=DEFAULT_SYSLOG_LEVEL,
                socktype=None,
                logformat=None,
            ):

            if logformat is None:
                logformat = DEFAULT_SYSLOG_FORMAT

            if default_level not in SYSLOG_LEVEL_MAP.keys():
                raise LoggerError('Unsupported syslog level value')

            logger = self.__get_or_create_logger__(name)
            handler = logging.handlers.SysLogHandler(address, facility, socktype)
            handler.level = default_level
            if not self.__match_handlers__(logger.handlers, handler):
                handler.setFormatter(logging.Formatter(logformat))
                logger.addHandler(handler)
                logger.setLevel(self.loglevel)

        def register_http_handler(self, name, url, method='POST'):
            logger = self.__get_or_create_logger__(name)
            try:
                host, path = urllib.splithost(url[url.index(':')+1:])
            except IndexError, emsg:
                raise LoggerError('Error parsing URL {0}: {1}'.format(url, emsg))

            handler = logging.handlers.HTTPHandler(host, url, method)
            if not self.__match_handlers__(logger.handlers, handler):
                logger.addHandler(handler)
                logger.setLevel(self.loglevel)

        def register_file_handler(self, name, directory,
                         logformat=None,
                         timeformat=None,
                         maxBytes=DEFAULT_LOGSIZE_LIMIT,
                         backupCount=DEFAULT_LOG_BACKUPS):
            """
            Register a common log file handler for rotating file based logs
            """

            if logformat is None:
                logformat = DEFAULT_LOGFILEFORMAT
            if timeformat is None:
                timeformat = DEFAULT_TIME_FORMAT

            if not os.path.isdir(directory):
                try:
                    os.makedirs(directory)
                except OSError:
                    raise LoggerError('Error creating directory: {0}'.format(directory))
            logfile = os.path.join(directory, '{0}.log'.format(name))

            logger = self.__get_or_create_logger__(name)
            handler = logging.handlers.RotatingFileHandler(
                filename=logfile,
                mode='a+',
                maxBytes=maxBytes,
                backupCount=backupCount
            )
            if not self.__match_handlers__(logger.handlers, handler):
                handler.setFormatter(logging.Formatter(logformat, timeformat))
                logger.addHandler(handler)
                logger.setLevel(self.loglevel)


        @property
        def level(self):
            return self._level
        @level.setter
        def level(self, value):
            if not isinstance(value, int):
                if value in LOGGING_LEVEL_NAMES:
                    value = getattr(logging, value)
                try:
                    value = int(value)
                    if value not in SYSLOG_LEVEL_MAP.values():
                        raise ValueError
                except ValueError:
                    raise ValueError('Invalid logging level value: {0}'.format(value))

            for logger in self.values():
                if hasattr(logger, 'setLevel'):
                    logger.setLevel(value)
            self._level = value

        # Compatibility for old API
        @property
        def loglevel(self):
            return self.level
        @loglevel.setter
        def loglevel(self, value):
            self.level = value

        # compatibility for old API
        def set_level(self, value):
            self.level = value

    def __getattr__(self, attr):
        return getattr(self.__instances[self.name], attr)

    def __setattr__(self, attr, value):
        setattr(self.__instances[self.name], attr, value)

    def __getitem__(self, item):
        return self.__instances[self.name][item]

    def __setitem__(self, item, value):
        self.__instances[self.name][item] = value



# coding=utf-8
"""Soundforest configuration database

Soundforest configuration database, implementing the database
classes in soundforest.models for cli scripts.

"""

import os
import hashlib

from soundforest import models, TreeError, SoundforestError
from soundforest.log import SoundforestLogger
from soundforest.defaults import DEFAULT_CODECS, DEFAULT_TREE_TYPES

FIELD_CONVERT_MAP = {
    'threads': lambda x: int(x)
}

class ConfigDB(object):
    """ConfigDB

    Configuration database settings API

    """

    __db_instance = None
    def __init__(self, path=None):
        if not ConfigDB.__db_instance:
            ConfigDB.__db_instance = ConfigDB.ConfigInstance(path)
        self.__dict__['ConfigDB.__db_instance'] = ConfigDB.__db_instance

    def __getattr__(self, attr):
        return getattr(self.__db_instance, attr)

    class ConfigInstance(models.SoundforestDB):
        """Configuration database instance

        Singleton instance of configuration database

        """
        def __init__(self, path):
            models.SoundforestDB.__init__(self, path=path)

            self.log = SoundforestLogger().default_stream

            treetypes = self.session.query(models.TreeTypeModel).all()
            if not treetypes:
                treetypes = []

                for name,description in DEFAULT_TREE_TYPES.items():
                    treetypes.append(models.TreeTypeModel(name=name, description=description))

                self.add(treetypes)
                self.commit()

            self.codecs = CodecConfiguration(db=self)
            self.sync = SyncConfiguration(db=self)

        def get(self, key):
            entry = self.session.query(models.SettingModel).filter(models.SettingModel.key==key).first()

            return entry is not None and entry.value or None

        def set(self, key, value):
            existing = self.session.query(models.SettingModel).filter(models.SettingModel.key==key).first()
            if existing:
                self.session.delete(existing)

            self.session.add(models.SettingModel(key=key, value=value))
            self.session.commit()

        def __getitem__(self, key):
            value = self.get(key)
            if value is None:
                raise KeyError

            return value

        def __setitem__(self, key, value):
            self.set(key, value)

        def __format_item__(self, key, value):
            if key in FIELD_CONVERT_MAP.keys():
                try:
                    value = FIELD_CONVERT_MAP[key](value)
                except ValueError:
                    raise SoundforestError('Invalid data in configuration for field {0}'.format(key))

            return value

        def has_key(self, key):
            return self.get(key) is not None

        def keys(self):
            return [s.key for s in self.session.query(models.SettingModel).all()]

        def items(self):
            return [(s.key, s.value) for s in self.session.query(models.SettingModel).all()]

        def values(self):
            return [s.value for s in self.session.query(models.SettingModel).all()]

    def update_tree(self, tree, update_checksum=True, progresslog=False):
        """
        Update tracks in database from loaded tree instance
        """
        added, updated, deleted, errors = 0, 0, 0, 0

        db_tree = self.get_tree(tree.path)
        albums = tree.as_albums()
        album_paths = [a.path for a in albums]
        track_paths = tree.realpaths

        processed = 0

        self.log.debug('{0} update tree'.format(tree.path))
        for album in albums:

            db_album = self.query(models.AlbumModel).filter(
                models.AlbumModel.tree == db_tree,
                models.AlbumModel.directory == album.path
            ).first()

            if db_album is None:
                self.log.debug('{0} add album {1}'.format(tree.path, album.path))
                db_album = models.AlbumModel(
                    tree=db_tree,
                    directory=album.path,
                    mtime=album.mtime
                )
                self.add(db_album)

            elif db_album.mtime != album.mtime:
                self.log.debug('{0} update album mtime {1}'.format(tree.path, album.relative_path()))
                db_album.mtime = album.mtime

            self.log.debug('{0} update album tracks {1}'.format(tree.path, album.relative_path()))
            for track in album:
                db_track = self.query(models.TrackModel).filter(
                    models.TrackModel.directory == track.path.directory,
                    models.TrackModel.filename == track.path.filename,
                ).first()

                if db_track is None:
                    self.log.debug('{0} add track {1}'.format(tree.path, track.relative_path()))
                    db_track = models.TrackModel(
                        tree=db_tree,
                        album=db_album,
                        directory=track.directory,
                        filename=track.filename,
                        extension=track.extension,
                        mtime=track.mtime,
                        deleted=False,
                    )
                    if self.update_track(track, update_checksum=update_checksum):
                        added +=1
                    else:
                        errors +=1

                elif db_track.mtime != track.mtime:
                    self.log.debug('{0} update track {1}'.format(tree.path, track.relative_path()))
                    if self.update_track(track, update_checksum=update_checksum):
                        updated += 1
                    else:
                        errors +=1

                elif not db_track.checksum and update_checksum:
                    self.log.debug('{0} update checksum {1}'.format(tree.path, track.relative_path()))
                    if self.update_track_checksum(track) is not None:
                        updated += 1
                    else:
                        errors +=1

                processed += 1
                if progresslog and processed % 1000 == 0:
                    self.log.debug('{0} processed {1:d} tracks'.format(tree.path, processed))

            self.commit()

        self.log.debug('{0} check for removed albums'.format(tree.path))
        for album in db_tree.albums:
            if album.path in album_paths or album.exists:
                continue

            self.log.debug('{0} remove album {1}'.format(tree.path, album.relative_path()))
            self.delete(album)

        self.log.debug('{0} check for removed tracks'.format(tree.path))
        for track in db_tree.tracks:
            if track.path in track_paths or track.exists:
                continue

            self.log.debug('{0} remove track {1}'.format(tree.path, track.relative_path()))
            self.delete(track)
            deleted += 1

        self.commit()

        if errors > 0:
            self.log.debug('Total {0:d} errors updating tree'.format(errors))

        return added, updated, deleted, processed, errors

    def find_tracks(self, path):
        track = self.get_track(path)
        if track is not None:
            return [track]

        db_tree = self.get_tree(path)
        if db_tree is not None:
            return db_tree.tracks

        db_album = self.get_album(path)
        if db_album is not None:
            return db_album.tracks

        if os.path.isdir(os.path.realpath(path)):
            return self.match_tracks_by_tree_prefix(path)

        return []

    def update_track(self, track, update_checksum=False):
        db_track = self.get_track(track.path)
        db_track.mtime = track.mtime

        oldtags = self.query(models.TrackModel).filter(models.TagModel.track == db_track)
        for tag in oldtags:
            self.delete(tag)

        try:
            tags = track.tags
        except TreeError, emsg:
            self.log.debug('ERROR loading {0}: {1}'.format(track.path, emsg))
            return False

        for tag, value in tags.items():
            if isinstance(value, list):
                value = value[0]
            self.add(models.TagModel(track=db_track, tag=tag, value=value))
        self.commit()

        if update_checksum:
            if self.update_track_checksum(track) is None:
                return False

        return True

    def update_track_checksum(self, track):
        db_track = self.get_track(track.path)
        if db_track is not None:
            checksum = track.checksum
            if db_track.checksum != checksum:
                db_track.checksum = checksum
                self.commit()
                return checksum
            else:
                return None
        else:
            return None

class ConfigDBDictionary(dict):
    """Configuration database dictionary

    Generic dictionary like instance of configuration databases

    """
    def __init__(self, db):
        self.log = SoundforestLogger().default_stream
        self.db = db

    def keys(self):
        return sorted(dict.keys(self))

    def items(self):
        return [(k, self[k]) for k in self.keys()]

    def values(self):
        return [self[k] for k in self.keys()]


class SyncConfiguration(ConfigDBDictionary):
    """SyncConfiguration

    Directory synchronization target configuration API

    """

    def __init__(self, db):
        super(SyncConfiguration, self).__init__(db)

        for target in self.db.registered_sync_targets:
            self[target.name] = target.as_dict()

    @property
    def threads(self):
        return int(self.db.get('threads'))

    @property
    def default_targets(self):
        return [k for k in self.keys() if self[k]['defaults']]

    def create_sync_target(self, name, synctype, src, dst, flags=None, defaults=False):
        self[name] = self.db.register_sync_target(name, synctype, src, dst, flags, defaults)


class CodecConfiguration(ConfigDBDictionary):

    """CodecConfiguration

    Audio codec decoder/encoder commands configuration API

    """

    def __init__(self, db):
        super(CodecConfiguration, self).__init__(db)

        self.load()

    def load(self):
        for codec in self.db.registered_codecs:
            self[str(codec.name)] = codec

        for name, settings in DEFAULT_CODECS.items():
            if name  in self.keys():
                continue

            codec = self.db.register_codec(name, **settings)
            self[str(codec.name)] = codec

    def extensions(self, codec):
        if codec in self.keys():
            return [codec] + [e.extension for e in self[codec].extensions]
        return []

    def match(self, path):
        ext = os.path.splitext(path)[1][1:]

        if ext == '':
            ext = path

        if ext in self.keys():
            return self[ext]

        for codec in self.values():
            if ext in [e.extension for e in codec.extensions]:
                return codec

        return None



# coding=utf-8
"""CLI utilities

Command line utilities for soundforest

"""

import sys
import os
import fnmatch
import time
import logging
import argparse
import tempfile
import signal
import socket
import threading
import subprocess

from setproctitle import setproctitle
from soundforest.config import ConfigDB
from soundforest.log import SoundforestLogger

def xterm_title(value, max_length=74, bypass_term_check=False):
    """
    Set title in xterm titlebar to given value, clip the title text to
    max_length characters.
    """
    TERM=os.getenv('TERM')
    TERM_TITLE_SUPPORTED = [ 'xterm', 'xterm-debian']
    if not bypass_term_check and TERM not in TERM_TITLE_SUPPORTED:
        return
    sys.stderr.write('\033]2;'+value[:max_length]+'',)
    sys.stderr.flush()

class ScriptError(Exception):
    pass

class ScriptThread(threading.Thread):
    """
    Common script thread base class
    """
    def __init__(self, name):
        threading.Thread.__init__(self)
        self.log = SoundforestLogger().default_stream
        self.status = 'not running'
        self.setDaemon(True)
        self.setName(name)

    def execute(self, command):
        p = subprocess.Popen(command, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
        return p.wait()

class ScriptThreadManager(list):
    def __init__(self, name, threads=None):
        self.log = SoundforestLogger().default_stream
        self.db = ConfigDB()
        if threads is None:
            threads = self.db.get('threads')
            if threads is None:
                threads = 1
        else:
            threads = int(threads)
        self.threads = threads

    def get_entry_handler(self, entry):
        raise NotImplementedError('Must be implemented in child class')

    def run(self):
        raise NotImplementedError('Must be implemented in child class')


class Script(object):
    """
    Common CLI tool setup class
    """
    def __init__(self, name=None, description=None, epilog=None, debug_flag=True):
        self.db = ConfigDB()
        self.name = os.path.basename(sys.argv[0])

        reload(sys)
        sys.setdefaultencoding('utf-8')

        setproctitle('%s %s' % (self.name, ' '.join(sys.argv[1:])))
        signal.signal(signal.SIGINT, self.SIGINT)

        if name is None:
            name = self.name

        self.logger = SoundforestLogger()
        self.log = self.logger.default_stream

        self.parser = argparse.ArgumentParser(
            prog=name,
            description=description,
            epilog=epilog,
            add_help=True,
            conflict_handler='resolve',
        )
        self.subcommand_parser = None
        self.subcommands = None

        if debug_flag:
            self.parser.add_argument('--debug', action='store_true', help='Show debug messages')

    def SIGINT(self, signum, frame):
        """
        Parse SIGINT signal by quitting the program cleanly with exit code 1
        """
        for t in filter(lambda t: t.name!='MainThread', threading.enumerate()):
            t.join()
        self.exit(1)

    def wait(self, poll_interval=1):
        """
        Wait for running threads to finish.
        Poll interval is time to wait between checks for threads
        """
        while True:
            active = filter(lambda t: t.name!='MainThread', threading.enumerate())
            if not len(active):
                break
            time.sleep(poll_interval)

    def exit(self, value=0, message=None):
        """
        Exit the script with given exit value.
        If message is not None, it is printed on screen.
        """
        if message is not None:
            self.message(message)

        while True:
            active = filter(lambda t: t.name!='MainThread', threading.enumerate())
            if not len(active):
                break
            time.sleep(1)
        sys.exit(value)

    def message(self, message):
        sys.stdout.write('%s\n' % message)
        sys.stdout.flush()

    def error(self, message):
        sys.stderr.write('%s\n' % message)
        sys.stdout.flush()

    def add_subcommand(self, command):
        if self.subcommand_parser is None:
            self.subcommand_parser = self.parser.add_subparsers(
                dest='command',
                help='Please select one command mode below',
                title='Command modes'
            )
            self.subcommands = {}

        if not isinstance(command, ScriptCommand):
            raise ScriptError('Subcommand must be a ScriptCommand instance')

        parser = self.subcommand_parser.add_parser(
            command.name,
            help=command.description,
            description=command.description,
            epilog=command.epilog
        )
        self.subcommands[command.name] = command
        command.script = self

        return parser

    def add_argument(self, *args, **kwargs):
        """
        Shortcut to add argument to main argumentparser instance
        """
        self.parser.add_argument(*args, **kwargs)

    def parse_args(self):
        """
        Call parse_args for parser and check for default logging flags
        """
        args = self.parser.parse_args()
        if hasattr(args, 'debug') and getattr(args, 'debug'):
            self.logger.set_level('DEBUG')
        return args

    def run(self):
        args = self.parse_args()
        if self.subcommand_parser is not None:
            self.subcommands[args.command].run(args)


class ScriptCommand(argparse.ArgumentParser):
    """
    Parent class for cli subcommands
    """
    def __init__(self, name, description='', epilog='', mode_flags=[]):
        self.script =  None
        self.name = name
        self.description = description
        self.epilog = epilog

        self.logger = SoundforestLogger()
        self.log = self.logger.default_stream

        if not isinstance(mode_flags, list):
            raise ScriptError('Mode flags must be a list')

        self.mode_flags = mode_flags
        self.selected_mode_flags = []

    @property
    def db(self):
        return self.script.db

    def exit(self, *args, **kwargs):
        self.script.exit(*args, **kwargs)

    def message(self, *args, **kwargs):
        self.script.message(*args, **kwargs)

    def parse_args(self, args):
        if hasattr(args, 'debug') and getattr(args, 'debug'):
            self.logger.set_level('DEBUG')

        self.selected_mode_flags = []
        for flag in self.mode_flags:
            if not hasattr(args, flag):
                continue

            if getattr(args, flag) not in ( None, False, [], ):
                self.selected_mode_flags.append(flag)

        xterm_title('soundforest %s' % (self.name))

        return args

    def match_path(self, path, matches=[]):
        for m in matches:
            m_realpath = os.path.realpath(m)
            t_realpath = os.path.realpath(path)

            if fnmatch.fnmatch(path, m):
                return True

            if path[:len(m)] == m:
                return True

            if t_realpath[:len(m_realpath)] == m_realpath:
                return True

        return False



# coding=utf-8
"""mp3 tags

mp3 file tag parser

"""

from mutagen.mp3 import MP3
from mutagen.easyid3 import EasyID3
from mutagen.id3 import ID3, APIC, ID3NoHeaderError
from mutagen.id3 import error as ID3Error

from soundforest.tags import TagError
from soundforest.tags.tagparser import TagParser, TrackNumberingTag, TrackAlbumart
from soundforest.tags.albumart import AlbumArt, AlbumArtError

MP3_ALBUMART_TAG = ''
MP3_ALBUMART_PIL_FORMAT_MAP = {
    # TODO - check these values
    'JPEG':     'image/jpeg',
    'PNG':      'image/png'
}

MP3_STANDARD_TAGS = {
    'album_artist':         ['TIT1'],
    'artist':               ['TPE1'],
    'composer':             ['TCOM'],
    'conductor':            ['TPE3'],
    'orchestra':            ['TPE2'],
    'performers':           ['TMCL'],
    'album':                ['TALB'],
    'title':                ['TIT2'],
    'genre':                ['TCON'],
    'comment':              ["COMM::'eng'"],
    'note':                 ['TXXX'],
    'description':          ['TIT3'],
    'year':                 ['TDRC'],
    'bpm':                  ['TBPM'],
    'label':                ['TPUB'],
    'copyright':            ['WCOP'],
    'license':              ['TOWN'],
    'sort_artist':          ['TSOP'],
    'sort_album':           ['TSOA'],
    'sort_title':           ['TSOT'],
}

MP3_TAG_FORMATTERS = {

}

def encode_frame(tag,value):
    """
    Return a mp3 frame object matching tag
    """
    try:
        m = __import__('mutagen.id3', globals(), {}, tag)
        tagclass = getattr(m, tag)
        if tagclass is None:
            raise AttributeError
    except AttributeError, emsg:
        raise TagError('Error importing ID3 frame {0}: {1}'.format(tag, emsg))
    if not isinstance(value, list):
        value = [value]
    return tagclass(encoding=3, text=value)

class MP3AlbumArt(TrackAlbumart):
    """
    Encoding of mp3 albumart to APIC tags
    """
    def __init__(self, track):
        if not isinstance(track, mp3):
            raise TagError('Track is not instance of mp3')

        super(MP3AlbumArt, self).__init__(track)

        try:
            self.tag = filter(lambda k:
                k[:5]=='APIC:',
                self.track.entry.keys()
            )[0]
        except IndexError:
            self.tag = None
            return

        try:
            albumart = AlbumArt()
            albumart.import_data(self.track.entry[self.tag].data)
        except AlbumArtError, emsg:
            raise TagError('Error reading mp3 albumart tag: {0}'.format(emsg))
        self.albumart = albumart

    def import_albumart(self, albumart):
        """
        Imports albumart object to the file tags.

        Sets self.track.modified to True
        """
        super(MP3AlbumArt, self).import_albumart(albumart)
        frame = APIC(0, albumart.mimetype, 0, '', albumart.dump())
        self.track.entry.tags.add(frame)
        self.track.modified = True

class MP3NumberingTag(TrackNumberingTag):
    """
    mp3 track numbering field handling
    """
    def __init__(self, track, tag):
        super(MP3NumberingTag, self).__init__(track, tag)

        if not self.track.entry.has_key(self.tag):
            return

        try:
            self.value, self.total = self.track.entry[self.tag].text[0].split('/')
        except ValueError:
            self.value = self.track.entry[self.tag].text[0]
            self.total = self.value

        try:
            self.value = int(self.value)
            self.total = int(self.total)
        except ValueError:
            raise TagError('Unsupported tag value for {0}: {1}'.format(
                self.tag,
                self.track.entry[self.tag].text[0]),
            )

    def save_tag(self):
        """
        Export this numbering information back to mp3 tags
        """
        if self.f_value is None:
            return
        value = self.value
        total = self.total
        if total is None:
            total = value
        if total < value:
            raise ValueError('Total is smaller than number')

        value = '{0:d}/{1:d}'.format(value, total)
        if self.track.entry.tags.has_key(self.tag):
            old_value = self.track.entry.tags[self.tag]
            if value == old_value:
                return
            del(self.track.entry.tags[self.tag])

        frame = encode_frame(self.tag, value)
        self.track.entry.tags.add(frame)
        self.track.modified = True

class mp3(TagParser):
    """
    Class for processing mp3 tags
    """
    def __init__(self, codec, path):
        super(mp3, self).__init__(codec, path, tag_map=MP3_STANDARD_TAGS)

        try:
            self.entry = MP3(self.path, ID3=ID3)
        except IOError:
            raise TagError('No ID3 header in {0}'.format(self.path))
        except ID3NoHeaderError:
            raise TagError('No ID3 header in {0}'.format(self.path))
        except RuntimeError:
            raise TagError('Runtime error loading {0}'.format(self.path))

        try:
            self.entry.add_tags()
        except ID3Error:
            pass

        self.albumart_obj = MP3AlbumArt(self)
        self.track_numbering = MP3NumberingTag(self, 'TRCK')
        self.disk_numbering = MP3NumberingTag(self, 'TPOS')

    def __getitem__(self, item):
        """
        Return tags formatted to unicode, decimal.Decimal or
        other supported types.
        Does not include albumart images, which are accessed
        by self.__getattr__('albumart')
        """
        if item == 'tracknumber':

            return [unicode('{0:d}'.format(self.track_numbering.value))]

        if item == 'totaltracks':
            return [unicode('{0:d}'.format(self.track_numbering.total))]

        if item == 'disknumber':
            return [unicode('{0:d}'.format(self.disk_numbering.value))]

        if item == 'totaldisks':
            return [unicode('{0:d}'.format(self.disk_numbering.total)]

        if item[:5] == 'APIC:':
            return self.albumart_obj

        fields = self.__tag2fields__(item)

        for field in fields:
            if not self.entry.has_key(field):
                continue

            tag = self.entry[field]
            if not isinstance(tag, list):
                tag = [tag]

            values = []
            for value in tag:
                matched = False

                for field in ['text', 'url']:
                    if hasattr(value, field):
                        value = getattr(value, field)
                        if isinstance(value, list):
                            value = value[0]

                        matched = True
                        break

                if value is None:
                    continue

                if not matched:
                    raise TagError('Error parsing {0}: {1}'.format(tag, dir(value)))

                if not isinstance(value, unicode):
                    try:
                        value = '{0:d}'.format(int(str(value)))
                    except ValueError, emsg:
                        pass

                    try:
                        value = unicode(value, 'utf-8')
                    except UnicodeDecodeError, emsg:
                        raise TagError('Error decoding {0} tag {1}: {2}'.format(self.path, field, emsg) )

                values.append(value)

            return values

        raise KeyError('No such tag: {0}'.format(fields))

    def __delitem__(self, item):
        tags = self.__tag2fields__(item)
        item = tags[0]
        for t in tags:
            if self.entry.tags.has_key(t):
                del(self.entry.tags[t])
                self.modified = True

    def keys(self):
        """
        Return tag names sorted with self.sort_keys()

        Itunes internal tags are ignored from results
        """
        keys = super(mp3, self).keys()
        if 'TRCK' in keys:
            keys.extend(['tracknumber', 'totaltracks'])
            keys.remove('TRCK')
        if 'TPOS' in keys:
            keys.extend(['disknumber', 'totaldisks'])
            keys.remove('TPOS')
        for k in keys:
            if k[:5] == 'APIC:':
                keys.remove(k)
        return self.sort_keys(keys)

    def set_tag(self, item, value):
        """
        Set given tag value to mp3
        """

        if item == 'tracknumber':
            self.track_numbering.value = value
            return

        if item == 'totaltracks':
            self.track_numbering.total = value
            return

        if item == 'disknumber':
            self.disk_numbering.value = value
            return

        if item == 'totaldisks':
            self.disk_numbering.total = value
            return

        if isinstance(value, list):
            value = value[0]

        tags = self.__tag2fields__(item)
        item = tags[0]
        for tag in tags[1:]:
            if self.entry.tags.has_key(tag):
                del(self.entry.tags[tag])

        if MP3_TAG_FORMATTERS.has_key(item):
            value = MP3_TAG_FORMATTERS[item](value)
        else:
            if not isinstance(value, unicode):
                value = unicode(value, 'utf-8')

        if self.entry.has_key(item):
            old_value = self.entry[item]
            if value == old_value:
                return
            del(self.entry.tags[item])

        frame = encode_frame(item, value)
        self.entry.tags.add(frame)
        self.modified = True

    def save(self):
        """
        Save mp3 tags to the file
        """
        super(mp3, self).save()



# coding=utf-8
"""Tag constants

Tag constants, including names and descriptions of standard tags

"""

import time

# These tags allow setting multiple valus in flac/vorbis
OGG_MULTIPLE_VALUES_TAGS = (
    'composer',
    'arranger',
    'lyricist',
    'author',
    'conductor',
    'performer',
    'ensemble',
    'part',
    'partnumber',
    'genre',
    'date',
    'location',
    'comment',
)


# Used for formatting output nicely
STANDARD_TAG_ORDER = [
    'album_artist',
    'composer',
    'artist',
    'conductor',
    'orchestra',
    'performers',
    'album',
    'title',
    'comment',
    'notes',
    'description',
    'location',
    'genre',
    'year',
    'bpm',
    'key',
    'tracknumber',
    'totaltracks',
    'disknumber',
    'totaldisks',
    'label',
    'license',
    'copyright',
    'sort_album_artist',
    'sort_composer',
    'sort_artist',
    'sort_performers',
    'sort_show',
    'sort_album',
    'sort_title',
]

# Tags common to all file formats. All tags in database are
# unicode strings or base64 formatted strings (if base64 flag
# is set)
STANDARD_TAG_MAP = {
    'album_artist': {
        'label': 'Album Artist',
        'description': 'Artist for the album',
    },
    'composer': {
        'label': 'Composer',
        'description': 'Composer of the track',
    },
    'artist': {
        'label': 'Artist',
        'description': 'Performer of the track',
    },
    'conductor': {
        'label': 'Conductor',
        'description': 'Conductor of the performance',
    },
    'orchestra': {
        'label': 'Orchestra',
        'description': 'Orchestra performing the track',
    },
    'performer': {
        'label': 'Performers',
        'description': 'Artists performing the track',
    },
    'album': {
        'label': 'Album',
        'description': 'Album title',
    },
    'title': {
        'label': 'Title',
        'description': 'Track title',
    },
    'genre': {
        'label': 'Genre',
        'description': 'Musical genre of the track',
    },
    'comment': {
        'label': 'Comment',
        'description': 'Comment for track',
    },
    'notes': {
        'label': 'Note',
        'description': 'A generic note for the track',
    },
    'description': {
        'label': 'Description',
        'description': 'Description of the item',
    },
    'location': {
        'label': 'Location',
        'description': 'Recording location of the track',
    },
    'year': {
        'label': 'Year',
        'description': 'Year the track was performed',
    },
    'bpm': {
        'label': 'BPM',
        'description': 'Track BPM value',
    },
    'key': {
        'label': 'Key',
        'description': 'Musical key of the track',
        },
    'tracknumber': {
        'label': 'Track Number',
        'description': 'Track number in album',
        },
    'totaltracks': {
        'label': 'Total Tracks',
        'description': 'Total number of tracks in album',
        },
    'disknumber': {
        'label': 'Disk Number',
        'description': 'Disk number in multi disk albums',
        },
    'totaldisks': {
        'label': 'Total Disks',
        'description': 'Total number of disks in album',
    },
    'label': {
        'label': 'Label',
        'description': 'Recording label for the album',
    },
    'license': {
        'label': 'License',
        'description': 'License for the track',
    },
    'copyright': {
        'label': 'Copyright',
        'description': 'Copyright owner for the track',
    },
    'sort_album_artist': {
        'label': 'Sort Album Artist',
        'description': 'Album artist name used for sorting',
    },
    'sort_artist': {
        'label': 'Sort Artist',
        'description': 'Artist name used for sorting',
    },
    'sort_composer': {
        'label': 'Sort Composer',
        'description': 'Composer name used for sorting',
    },
    'sort_performers': {
        'label': 'Sort Performers',
        'description': 'Performers used for sorting',
    },
    'sort_show': {
        'label': 'Sort Show',
        'description': 'Show name used for sorting',
    },
    'sort_album': {
        'label': 'Sort Album',
        'description': 'Album name used for sorting',
    },
    'sort_title': {
        'label': 'Sort Title',
        'description': 'Title used for sorting',
    },
}

DATE_FORMATS = [
    '%Y-%m-%dT%H:%M:%SZ',
    '%Y-%m-%d',
    '%Y%m%d',
    '%Y',
]

def sorted_tags(tags):
    sorted_tags = []

    for tag in STANDARD_TAG_ORDER:
        if tag in tags:
            sorted_tags.append(tag)

    for tag in tags:
        if tag in sorted(tags):
            if tag not in sorted_tags:
                sorted_tags.append(tag)

    return sorted_tags

def parsedate(value):
    tval = None
    for fmt in DATE_FORMATS:
        try:
            tval = time.strptime(value, fmt)
            break
        except ValueError, emsg:
            continue
    return tval




# coding=utf-8
"""vorbis tags

Vorbis file tag parser

"""

from mutagen.oggvorbis import OggVorbis, OggVorbisHeaderError

from soundforest.tags import TagError
from soundforest.tags.constants import OGG_MULTIPLE_VALUES_TAGS
from soundforest.tags.tagparser import TagParser, TrackNumberingTag, TrackAlbumart
from soundforest.tags.albumart import AlbumArt, AlbumArtError

VORBIS_ALBUMART_TAG = 'METADATA_BLOCK_PICTURE'

VORBIS_STANDARD_TAGS = {
    'album_artist':         ['ALBUM_ARTIST'],
    'artist':               ['ARTIST'],
    'arranger':             ['ARRANGER'],
    'author':               ['AUTHOR'],
    'composer':             ['COMPOSER'],
    'conductor':            ['CONDUCTOR'],
    'ensemble':             ['ENSEMBLE'],
    'orchestra':            ['ORCHESTRA'],
    'performer':            ['PERFORMER'],
    'publisher':            ['PUBLISHER'],
    'lyricist':             ['LYRICIST'],
    'album':                ['ALBUM'],
    'title':                ['TITLE'],
    'partnumber':           ['PARTNUMBER'],
    'tracknumber':          ['TRACKNUMBER'],
    'disknumber':           ['DISKNUMBER'],
    'genre':                ['GENRE'],
    'comment':              ['COMMENT'],
    'note':                 ['NOTE'],
    'description':          ['DESCRIPTION'],
    'location':             ['LOCATION'],
    'year':                 ['DATE'],
    'bpm':                  ['BPM'],
    'rating':               ['RATING'],
    'label':                ['LABEL'],
    'labelno':              ['LABELNO'],
    'opus':                 ['OPUS'],
    'isrc':                 ['ISRC'],
    'ean':                  ['EAN/UPN'],
    'lyrics':               ['LYRICS'],
    'website':              ['WEBSITE'],
    'copyright':            ['COPYRIGHT'],
    'version':              ['VERSION'],
    'sourcemedia':          ['SOURCEMEDIA'],
    'encoding':             ['ENCODING'],
    'encoded_by':           ['ENCODED-BY'],
    'sort_album_artist':    ['SORT_ALBUM_ARTIST'],
    'sort_artist':          ['SORT_ARTIST'],
    'sort_composer':        ['SORT_COMPOSER'],
    'sort_performer':       ['SORT_PERFORMER'],
    'sort_show':            ['SORT_SHOW'],
    'sort_album':           ['SORT_ALBUM'],
    'sort_title':           ['SORT_TITLE'],
}

VORBIS_REPLAYGAIN_TAGS = {
    'album_gain':           ['REPLAYGAIN_ALBUM_GAIN'],
    'album_peak':           ['REPLAYGAIN_ALBUM_PEAK'],
    'track_gain':           ['REPLAYGAIN_TRACK_GAIN'],
    'track_peak':           ['REPLAYGAIN_TRACK_PEAK'],
}

VORBIS_TAG_FORMATTERS = {

}

VORBIS_EXTRA_TAGS = {
}

class VorbisNumberingTag(TrackNumberingTag):
    """
    Vorbis tags for storing track or disk numbers.
    The tag can be either a single number or two numbers separated by /
    If total is given, the value must be integer.
    """
    def __init__(self, track, tag):
        super(VorbisNumberingTag, self).__init__(track, tag)

        if not isinstance(track, vorbis):
            raise TagError('Track is not instance of vorbis')

        if not self.track.entry.has_key(self.tag):
            return

        value = self.track.entry[self.tag]
        try:
            value, total = value[0].split('/', 1)
        except ValueError:
            total = None
        self.value = value
        self.total = total

    def save_tag(self):
        """
        Set new numbering information to vorbis tags, marking file
        dirty to require saving but not saving tags.
        """
        value = self.__repr__()
        if value is not None:
            self.track.entry[self.tag] = '{0}'.format(value)
            self.track.modified = True

class vorbis(TagParser):
    """
    Class for processing Ogg Vorbis file tags
    """
    def __init__(self, codec, path):
        super(vorbis, self).__init__(codec, path, tag_map=VORBIS_STANDARD_TAGS)

        try:
            self.entry = OggVorbis(path)
        except IOError, emsg:
            raise TagError('Error opening {0}: {1}'.format(path, str(emsg)))
        except OggVorbisHeaderError, emsg:
            raise TagError('Error opening {0}: {1}'.format(path, str(emsg)))

        self.albumart_obj = None
        self.track_numbering = VorbisNumberingTag(self, 'TRACKNUMBER')
        self.disk_numbering = VorbisNumberingTag(self, 'DISKNUMBER')

    def __getitem__(self, item):
        if item == 'tracknumber':
            return [unicode('{0:d}'.format(self.track_numbering.value))]
        if item == 'totaltracks':
            return [unicode('{0:d}'.format(self.track_numbering.total))]
        if item == 'disknumber':
            return [unicode('{0:d}'.format(self.disk_numbering.value))]
        if item == 'totaldisks':
            return [unicode('{0:d}'.format(self.disk_numbering.total))]

        super(vorbis, self).__getitem__(item)

    def __delitem__(self, item):
        try:
            item, value = item.split('=', 1)
        except ValueError:
            value = None

        fields = self.__tag2fields__(item)
        for tag in fields:
            tag = self.__field2tag__(tag)
            if not self.has_key(tag):
                continue

            if value is None:
                del self.entry[tag]
                self.modified = True

            elif value in self.entry[tag]:
                self.entry[tag] = [x for x in self.entry[tag] if x != value]

                if not self.entry[tag]:
                    del self.entry[tag]

                self.modified = True

    def __field2tag__(self, field):
        super(vorbis, self).__field2tag__(field.upper())

    def keys(self):
        """
        Return tag names sorted with self.sort_keys()
        """

        keys = super(vorbis, self).keys()

        if 'TOTALTRACKS' in keys:
            keys.remove('TOTALTRACKS')

        if 'TOTALDISKS' in keys:
            keys.remove('TOTALDISKS')

        if 'TRACKNUMBER' in [x.upper() for x in keys]:
            if self.track_numbering.total is not None:
                keys.append('totaltracks')

        if 'DISKNUMBER' in [x.upper() for x in keys]:
            if self.disk_numbering.total is not None:
                keys.append('totaldisks')

        if VORBIS_ALBUMART_TAG in [x.upper() for x in keys]:
            keys.remove(VORBIS_ALBUMART_TAG)

        for replaygain_tag_fields in VORBIS_REPLAYGAIN_TAGS.values():
            for tag in replaygain_tag_fields:
                if tag in keys:
                    keys.remove(tag)

        return [x for x in self.sort_keys(keys)]

    def has_key(self,  tag):
        return tag.lower() in self.keys()

    def set_tag(self, item, value):
        """
        All vorbis tags are unicode strings, and there can be multiple
        tags with same name.

        We do special precessing for track and disk numbering.
        """
        if item == 'tracknumber':
            self.track_numbering.value = value
            self.modified = True
            return

        if item == 'totaltracks':
            self.track_numbering.total = value
            self.modified = True
            return

        if item == 'disknumber':
            self.disk_numbering.value = value
            self.modified = True
            return

        if item == 'totaldisks':
            self.disk_numbering.total = value
            self.modified = True
            return

        if not isinstance(value, list):
            value = [value]

        value = [x for x in set(value)]

        tags = self.__tag2fields__(item)
        item = tags[0]

        for tag in tags:
            if self.entry.has_key(tag):
                if tag.lower() in OGG_MULTIPLE_VALUES_TAGS:
                    value = set(self.entry[tag] + value)

                del self.entry[tag]

        entries =[]
        for v in value:
            if VORBIS_TAG_FORMATTERS.has_key(item):
                entries.append(VORBIS_TAG_FORMATTERS[item](v))

            else:
                if not isinstance(v, unicode):
                    v = unicode(v, 'utf-8')
                entries.append(v)

        self.entry[item] = entries
        self.modified = True



# coding=utf-8
"""AAC tags

AAC file tag parser

"""

import base64
import struct

from mutagen.mp4 import MP4, MP4Cover, MP4StreamInfoError, MP4MetadataValueError

from soundforest.tags import TagError
from soundforest.tags.tagparser import TagParser, TrackNumberingTag, TrackAlbumart
from soundforest.tags.albumart import AlbumArt, AlbumArtError

# Albumart filed processing
AAC_ALBUMART_TAG = 'covr'
AAC_ALBUMART_PIL_FORMAT_MAP = {
    'JPEG':     MP4Cover.FORMAT_JPEG,
    'PNG':      MP4Cover.FORMAT_PNG
}

AAC_STANDARD_TAGS = {
    'album_artist':         ['aART'],
    'artist':               ['\xa9ART'],
    'composer':             ['\xa9wrt'],
    'conductor':            ['cond'],
    'description':          ['desc'],
    'discid':               ['disc'],
    'orchestra':            ['orch'],
    'performers':           ['ense'],
    'album':                ['\xa9alb'],
    'title':                ['\xa9nam'],
    'genre':                ['\xa9gen'],
    'comment':              ['\xa9cmt'],
    'lyrics':               ['\xa9lyr'],
    'note':                 ['note'],
    'description':          ['desc'],
    'location':             ['loca'],
    'year':                 ['\xa9day'],
    'bpm':                  ['tmpo'],
    'rating':               ['rati'],
    'replaygain':           ['repl'],
    'mcn':                  ['mcn\x00'],
    'label':                ['labe'],
    'copyright':            ['cprt'],
    'license':              ['lice'],
    'podcast_url':          ['purl'],
    'podcast_episode_guid': ['egid'],
    'podcast_category':     ['catg'],
    'podcast_keywords':     ['keyw'],
    'sort_album_artist':    ['soaa'],
    'sort_artist':          ['soar'],
    'sort_composer':        ['soco'],
    'sort_performers':      ['sopr'],
    'sort_show':            ['sosn'],
    'sort_album':           ['soal'],
    'sort_title':           ['sonm'],

}

# Internal program tags for itunes. Ignored by current code
ITUNES_TAG_MAP = {
    # Booleans
    'compilation':          ['cpil'],
    'gapless_album':        ['pgap'],
    'podcast':              ['pcst'],
    # iTunes grouping flag
    'grouping':             ['\xa9grp'],
    # Indicates the encoder command used to encode track
    'encoder':              ['\xa9too', 'enco'],
    # Another way iTunes stores tool info
    'itunes_tool':          ['----:com.apple.iTunes:tool'],
    # iTunes encoder and normalization data
    'itunes_encoder':       ['----:com.apple.iTunes:cdec'],
    'itunes_normalization': ['----:com.apple.iTunes:iTunNORM'],
    # XML info for song
    'itunes_movi':          ['----:com.apple.iTunes:iTunMOVI'],
    # NO idea what this is
    'itunes_smbp':          ['----:com.apple.iTunes:iTunSMPB'],
    # iTunes store purchase details
    'purchase_date':        ['purd'],
    'purchaser_email':      ['apID'],
    # Tags for video shows
    'video_show':           ['tvsh'],
    'video_episode':        ['tven'],
    # XID is internal itunes metadata reference
    'xid':                  ['xid'],
}

AAC_UNOFFICIAL_TAGS = {
    # Musicbrainz ID reference
    'musicbrainz_id':       ['musi'],
}

# These values are (value, value) pairs in metadata
AAC_INTEGER_TUPLE_TYPES = [ 'trkn', 'disk' ]

# Placeholder to write lambda functions to process specific tags if needed
AAC_TAG_FORMATTERS = {
    'tmpo':    lambda x: int(x),
}


class AACAlbumArt(TrackAlbumart):
    """
    Thin wrapper to process AAC object albumart files

    Technically supports setting albumart to other tags than
    standard AAC_ALBUMART_TAG. Don't do this, not tested.
    """
    def __init__(self, track, tag=AAC_ALBUMART_TAG):
        if not isinstance(track, aac):
            raise TagError('Track is not instance of aac')

        super(AACAlbumArt, self).__init__(track)

        self.tag = AAC_ALBUMART_TAG
        if not self.track.entry.has_key(self.tag):
            return

        try:
            albumart = AlbumArt()
            albumart.import_data(self.track.entry[self.tag][0])
        except AlbumArtError, emsg:
            raise TagError('Error reading AAC albumart tag: {0}'.format(emsg))

        self.albumart = albumart

    def import_albumart(self, albumart):
        """
        Imports albumart object to the file tags.

        Sets self.track.modified to True
        """
        super(AACAlbumArt, self).import_albumart(albumart)

        try:
            img_format = AAC_ALBUMART_PIL_FORMAT_MAP[self.albumart.get_fileformat()]
        except KeyError:
            raise TagError('Unsupported albumart format {0}'.format(self.albumart.get_fileformat() ))
        try:
            tag = MP4Cover(data=self.albumart.dump(), imageformat=img_format)
        except MP4MetadataValueError, emsg:
            raise TagError('Error encoding albumart: {0}'.format(emsg))

        if self.track.entry.has_key(self.tag):
            if self.track.entry[self.tag] != [tag]:
                del self.track.entry[self.tag]
            else:
                return False
        self.track.entry[self.tag] = [tag]
        self.track.modified = True
        return self.track.modified


class AACIntegerTuple(TrackNumberingTag):
    """
    AAC field for ('item', 'total items') type items in tags.
    Used for track and disk numbering
    """
    def __init__(self, track, tag):
        super(AACIntegerTuple, self).__init__(track, tag)

        if not self.track.entry.has_key(self.tag):
            return

        self.value, self.total = self.track.entry[self.tag][0]

    def delete_tag(self):
        self.value = None
        self.total = None
        if self.tag in self.track.entry.keys():
            del self.track.entry[self.tag]
            self.track.modified = True

    def save_tag(self):
        """
        Export this numbering information back to AAC tags.

        If value is None, ignore both values without setting tag.
        If total is None but value is set, set total==value.
        """
        if self.f_value is None:
            return
        value = self.value
        total = self.total
        if total is None:
            total = value
        if total < value:
            raise ValueError('Total is smaller than number')
        self.track.entry[self.tag] = [(value, total)]
        self.track.modified = True


class aac(TagParser):
    """
    Class for processing AAC file tags
    """
    def __init__(self, codec, path):
        super(aac, self).__init__(codec, path, tag_map=AAC_STANDARD_TAGS)

        try:
            self.entry = MP4(self.path)
        except IOError, emsg:
            raise TagError('Error opening {0}: {1}'.format(path, str(emsg)))

        except MP4StreamInfoError, emsg:
            raise TagError('Error opening {0}: {1}'.format(path, str(emsg)))

        except RuntimeError, emsg:
            raise TagError('Error opening {0}: {1}'.format(path, str(emsg)))

        except struct.error:
            raise TagError('Invalid tags in {0}'.format(path))

        self.supports_albumart = True
        self.albumart_obj = AACAlbumArt(self)
        self.track_numbering = AACIntegerTuple(self, 'trkn')
        self.disk_numbering = AACIntegerTuple(self, 'disk')

    def __getitem__(self, item):
        if item == 'tracknumber':
            return [unicode('{0:d}'.format(self.track_numbering.value))]

        if item == 'totaltracks':
            return [unicode('{0:d}'.format(self.track_numbering.total))]

        if item == 'disknumber':
            return [unicode('{0:d}'.format(self.disk_numbering.value))]

        if item == 'totaldisks':
            return [unicode('{0:d}'.format(self.disk_numbering.total))]

        if item == 'unknown_tags':
            keys = []
            for tag in self.entry.keys():
                if tag in AAC_STANDARD_TAGS.values():
                    continue
                if tag in ITUNES_TAG_MAP.values():
                    continue
                if tag in AAC_UNOFFICIAL_TAGS.values():
                    continue
                keys.append(tag)

            return keys

        return super(aac, self).__getitem__(item)

    def __delitem__(self, item):
        if item in ['tracknumber', 'totaltracks']:
            return self.track_numbering.delete_tag()

        if item in ['disknumber', 'totaldisks']:
            return self.disk_numbering.delete_tag()

        return super(aac, self).__delitem__(item)

    def set_tag(self, item, value):
        """
        Set given tag to correct type of value in tags.

        Normal tag values in AAC tags are always a list
        and you can pass this function a list to set all values.

        Tracknumber, totaltracks, disknumber and totaldisks
        attributes must be integers.

        Existing tag value list is replaced.
        """
        if item == 'tracknumber':
            self.track_numbering.value = value
        elif item == 'totaltracks':
            self.track_numbering.total = value
        elif item == 'disknumber':
            self.disk_numbering.value = value
        elif item == 'totaldisks':
            self.disk_numbering.total = value
        else:
            if not isinstance(value, list):
                value = [value]

            tags = self.__tag2fields__(item)
            item = tags[0]
            for tag in tags[1:]:
                if self.entry.has_key(tag):
                    del self.entry[tag]

            entries =[]
            for v in value:
                if AAC_TAG_FORMATTERS.has_key(item):
                    formatted = AAC_TAG_FORMATTERS[item](v)
                    entries.append(formatted)
                else:
                    if not isinstance(v, unicode):
                        v = unicode(v, 'utf-8')
                    entries.append(v)
            self.entry[item] = entries

        self.modified = True

    def keys(self):
        """
        Return tag names sorted with self.sort_keys()

        Itunes internal tags are ignored from results
        """
        keys = super(aac, self).keys()

        if 'trkn' in keys:
            keys.extend(['tracknumber', 'totaltracks'])
            keys.remove('trkn')
        if 'disk' in keys:
            keys.extend(['disknumber', 'totaldisks'])
            keys.remove('disk')
        if 'covr' in keys:
            keys.remove('covr')
        for itunes_tags in ITUNES_TAG_MAP.values():
            for tag in itunes_tags:
                if tag in keys:
                    keys.remove(tag)
        return self.sort_keys(keys)

    def has_key(self, key):
        """Extended has_key

        Extend has_key with tracknumber info

        """
        return key in self.keys()

    def save(self):
        """
        Save AAC tags to the file
        """
        try:
            super(aac, self).save()
        except MP4MetadataValueError, emsg:
            raise TagError(emsg)
        self.modified = False



# coding=utf-8
"""Music file formats

Guessing of supported file formats and codecs based on extensions

"""

import os
import tempfile

from subprocess import Popen, PIPE

from soundforest import normalized, SoundforestError, CommandPathCache
from soundforest.config import ConfigDB
from soundforest.defaults import SOUNDFOREST_CACHE_DIR
from soundforest.log import SoundforestLogger
from soundforest.metadata import Metadata

logger = SoundforestLogger().default_stream

TAG_PARSERS = {
    'm4a':      'soundforest.tags.formats.aac.aac',
    'm4r':      'soundforest.tags.formats.aac.aac',
    'mp3':      'soundforest.tags.formats.mp3.mp3',
    'flac':     'soundforest.tags.formats.flac.flac',
    'vorbis':   'soundforest.tags.formats.vorbis.vorbis',
}

PATH_CACHE = CommandPathCache()
PATH_CACHE.update()

db = ConfigDB()

def filter_available_command_list(commands):
    available = []
    for cmd in commands:
        try:
            executable = cmd.command.split(' ', 1)[0]
        except IndexError:
            executable = cmd.command
            pass
        if PATH_CACHE.which(executable) is None:
            continue
        available.append(cmd.command)

    return available

def match_codec(path):
    ext = os.path.splitext(path)[1][1:]

    if ext == '':
        ext = path

    if ext in db.codecs.keys():
        return db.codecs[ext]

    for codec in db.codecs.values():
        if ext in [e.extension for e in codec.extensions]:
            return codec

    return None

def match_metadata(path):
    metadata = Metadata()
    m = metadata.match(path)
    if not m:
        return None

    return m

class path_string(unicode):
    def __init__(self, path):
        if isinstance(path, unicode):
            super(path_string, self).__init__(normalized(path).encode('utf-8'))
        else:
            super(path_string, self).__init__(normalized(path))

    @property
    def exists(self):
        if os.path.isdir(self) or os.path.isfile(self):
            return True
        return False

    @property
    def isdir(self):
        return os.path.isdir(self)

    @property
    def isfile(self):
        return os.path.isfile(self)

    @property
    def no_ext(self):
        return os.path.splitext(self)[0]

    @property
    def directory(self):
        return os.path.dirname(self)

    @property
    def filename(self):
        return os.path.basename(self)

    @property
    def extension(self):
        return os.path.splitext(self)[1][1:]


class AudioFileFormat(object):
    """AudioFileFormat

    Common file format wrapper for various codecs

    """

    def __init__(self, path):
        self.log =  SoundforestLogger().default_stream
        self.path = path_string(path)
        self.codec = None
        self.description = None
        self.is_metadata = False

        self.codec = match_codec(path)
        if self.codec is not None:
            self.description = self.codec.description.lower()

        else:
            m = match_metadata(path)
            if m:
                self.is_metadata = True
                self.description = m.description.lower()

            elif os.path.isdir(path):
                self.description = 'unknown directory'

            else:
                self.description = 'unknown file format'

    def __repr__(self):
        return '{0} {1}'.format(self.codec, self.path)

    @property
    def directory(self):
        return os.path.dirname(self.path)

    @property
    def filename(self):
        return os.path.basename(self.path)

    @property
    def extension(self):
        return os.path.splitext(self.path)[1][1:]

    @property
    def size(self):
        if not self.path.isfile:
            return None
        return os.stat(self.path).st_size

    @property
    def ctime(self):
        if not self.path.isfile:
            return None
        return os.stat(self.path).st_ctime

    @property
    def mtime(self):
        if not self.path.isfile:
            return None
        return os.stat(self.path).st_mtime

    def get_temporary_file(self, dir=SOUNDFOREST_CACHE_DIR, prefix='tmp', suffix=''):
        if not os.path.isdir(dir):
            try:
                os.makedirs(dir)
            except IOError, (ecode, emsg):
                raise SoundforestError('Error creating directory {0}: {1}'.format(SOUNDFOREST_CACHE_DIR, emsg))
            except OSError, (ecode, emsg):
                raise SoundforestError('Error creating directory {0}: {1}'.format(SOUNDFOREST_CACHE_DIR, emsg))

        return tempfile.mktemp(dir=dir, prefix=prefix, suffix=suffix)

    def get_tag_parser(self):
        if self.codec is None or self.codec.name not in TAG_PARSERS.keys():
            return None

        try:
            classpath = TAG_PARSERS[self.codec.name]
            module_path = '.'.join(classpath.split('.')[:-1])
            class_name = classpath.split('.')[-1]
            m = __import__(module_path, globals(), fromlist=[class_name])

        except KeyError, emsg:
            return None

        return getattr(m, class_name)

    def get_available_encoders(self):
        if self.codec is None or not self.codec.encoders:
            return []

        return filter_available_command_list(self.codec.encoders)

    def get_available_decoders(self):
        if self.codec is None or not self.codec.decoders:
            return []

        return filter_available_command_list(self.codec.decoders)

    def get_available_testers(self):
        if self.codec is None or not self.codec.testers:
            return []

        return filter_available_command_list(self.codec.testers)

    def execute(self, args):
        self.log.debug('running: {0}'.format(' '.join(args)))
        p = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        (stdout, stderr) = p.communicate()

        if stdout:
            self.log.debug('output:\n{0}'.format(stdout))
        if stderr:
            self.log.debug('errors:\n{0}'.format(stderr))

        return p.returncode, stdout, stderr



"""
Setup for soundforest package for setuptools
"""

import glob
from setuptools import setup, find_packages
from soundforest import __version__

setup(
    name = 'soundforest',
    keywords = 'Sound Audio File Tree Codec Database',
    description = 'Audio file library manager',
    version = __version__,
    author = 'Ilkka Tuohela',
    author_email = 'hile@iki.fi',
    license = 'PSF',
    url = 'https://github.com/hile/soundforest',
    packages = find_packages(),
    scripts = glob.glob('bin/*'),
    install_requires = (
        'sqlalchemy>=1.0.11',
        'setproctitle',
        'requests',
        'lxml',
        'pytz',
        'mutagen',
        'pillow',
    ),
)
